// Package constants is responsible for defining private constants used
// by the common module of Ubuntu Insights.
package constants

import (
	"log/slog"
)

const (
	// DefaultLogLevel is the default log level selected without any verbosity flags.
	DefaultLogLevel = slog.LevelWarn
)
